@@ -3,10 +3,10 @@
  *
  * Author:	Miquel van Smoorenburg, miquels@cistron.nl
  * 
- * Version:	@(#)dowall.c  2.73  25-Nov-1997  miquels@cistron.nl
+ * Version:	@(#)dowall.c  2.75  28-Jul-1998  miquels@cistron.nl
  *
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-1997 Miquel van Smoorenburg.
+ *		Copyright 1991-1998 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
@@ -41,6 +41,26 @@ int arg;
 	longjmp(jbuf, 1);
 }
 
+
+/*
+ *	Print a text, escape all characters not in Latin-1.
+ */
+void feputs(char *line, FILE *fp)
+{
+	unsigned char *p;
+
+	for (p = (unsigned char *)line; *p; p++) {
+		if (strchr("\t\r\n", *p) ||
+		    (*p >= 32 && *p <= 127) || (*p >= 160)) {
+			fputc(*p, fp);
+		} else {
+			fprintf(fp, "^%c", (*p & 0x1f) + 'A' - 1);
+		}
+	}
+	fflush(fp);
+}
+
+
 /*
  *	Wall function.
  */
@@ -91,7 +111,7 @@ void wall(char *text, int fromshutdown, int remote)
 			"\007\r\nRemote broadcast message %s...\r\n\r\n",
 			date);
 	} else {
-		sprintf(line,
+		snprintf(line, sizeof(line),
 			"\007\r\nBroadcast message from %s %s%s...\r\n\r\n",
 			user, ttynm, date);
 	}
@@ -106,7 +126,9 @@ void wall(char *text, int fromshutdown, int remote)
 	while ((utmp = getutent()) != NULL) {
 		if(utmp->ut_type != USER_PROCESS ||
 		   utmp->ut_user[0] == 0) continue;
-		sprintf(term, "/dev/%s", utmp->ut_line);
+		if ((p = strrchr(utmp->ut_line, '/')) == 0)
+			p = utmp->ut_line;
+		sprintf(term, "/dev/%s", p);
 	
 		/*
 		 *	Sometimes the open/write hangs in spite of the O_NDELAY
@@ -120,7 +142,7 @@ void wall(char *text, int fromshutdown, int remote)
 		    (fd = open(term, O_WRONLY | O_NDELAY )) > 0) {
 			if ((tp = fdopen(fd, "w")) != NULL) {
 				fputs(line, tp);
-				fputs(text, tp);
+				feputs(text, tp);
 #ifdef AEROSMITH
 				if (fromshutdown &&
 				    !strcmp(utmp->ut_user, "tyler"))
@@ -137,7 +159,7 @@ void wall(char *text, int fromshutdown, int remote)
 #else
 		if (setjmp(jbuf) == 0 && (tp = fopen(term, "w")) != NULL) {
 			fputs(line, tp);
-			fputs(text, tp);
+			feputs(text, tp);
 			alarm(0);
 			fclose(tp);
 			tp = NULL;
