@@ -5,7 +5,7 @@
  *		     init [0123456SsQqAaBbCc]
  *		  telinit [0123456SsQqAaBbCc]
  *
- * Version:	@(#)init.c  2.75-2  02-Jun-1998  miquels@cistron.nl
+ * Version:	@(#)init.c  2.76-1  03-Nov-1998  miquels@cistron.nl
  *
  *		This file is part of the sysvinit suite,
  *		Copyright 1991-1998 Miquel van Smoorenburg.
@@ -61,9 +61,9 @@
 				sigaction(sig, &sa, NULL);
 
 /* Version information */
-char *Version = "@(#) init 2.75-2  02-Jun-1998  miquels@cistron.nl";
-char *bootmsg = "version 2.75 booting";
-#define E_VERSION "INIT_VERSION=sysvinit-2.75"
+char *Version = "@(#) init 2.76-1  03-Nov-1998  miquels@cistron.nl";
+char *bootmsg = "version 2.76 booting";
+#define E_VERSION "INIT_VERSION=sysvinit-2.76"
 
 CHILD *family = NULL;		/* The linked list of all entries */
 CHILD *newFamily = NULL;	/* The list after inittab re-read */
@@ -90,7 +90,7 @@ int sltime = 5;			/* Sleep time between TERM and KILL */
 char *argv0;			/* First arguments; show up in ps listing */
 int maxproclen;			/* Maximal length of argv[0] */
 struct utmp utproto;		/* Only used for sizeof(utproto.ut_id) */
-char *console_dev = CONSOLE;	/* Console device. */
+char *console_dev;		/* Console device. */
 int wrote_iosave = 0;		/* Did we write ioctl.save yet? */
 int pipe_fd = -1;		/* /dev/initctl */
 int did_boot = 0;		/* Did we already do BOOT* stuff? */
@@ -166,35 +166,20 @@ struct {
 };
 
 /*
- *	Sleep a number of seconds
+ *	Sleep a number of seconds.
+ *
+ *	This only works correctly because the linux select updates
+ *	the elapsed time in the struct timeval passed to select!
  */
 void do_sleep(int sec)
 {
-	int old_alrm;
-	int f;
-	int ga;
+	struct timeval tv;
 
-	ga = ISMEMBER(got_signals, SIGALRM);
+	tv.tv_sec = sec;
+	tv.tv_usec = 0;
 
-	/*
-	 *	Save old alarm time and wait 'sec' seconds.
-	 */
-	old_alrm = alarm(0);
-	for(f = 0; f < sec; f++) {
-		DELSET(got_signals, SIGALRM);
-		alarm(1);
-		while(ISMEMBER(got_signals, SIGALRM) == 0) pause();
-	}
-
-	/*
-	 *	Reset old values of got_alrm flag and the timer
-	 */
-	if (ga)
-  		ADDSET(got_signals, SIGALRM);
-	else
-		DELSET(got_signals, SIGALRM);
-	if (old_alrm)
-		alarm(old_alrm - sec > 0 ? old_alrm - sec : 1);
+	while(select(0, NULL, NULL, NULL, &tv) < 0 && errno == EINTR)
+		;
 }
 
 /*
@@ -428,9 +413,46 @@ int setproctitle(char *fmt, ...)
 }
 
 /*
- *	Open a device with retries.
+ *	Reset console tries.
  */
-int serial_open(char *dev, int mode)
+void console_init(void)
+{
+	int fd;
+	int tried_devcons = 0;
+	int tried_vtmaster = 0;
+	char *s;
+
+	if ((s = getenv("CONSOLE")) != NULL)
+		console_dev = s;
+	else {
+		console_dev = CONSOLE;
+		tried_devcons++;
+	}
+
+	while ((fd = open(console_dev, O_RDONLY|O_NONBLOCK)) < 0) {
+		if (!tried_devcons) {
+			tried_devcons++;
+			console_dev = CONSOLE;
+			continue;
+		}
+		if (!tried_vtmaster) {
+			tried_vtmaster++;
+			console_dev = VT_MASTER;
+			continue;
+		}
+		break;
+	}
+	if (fd < 0)
+		console_dev = "/dev/null";
+	else
+		close(fd);
+}
+
+
+/*
+ *	Open the console with retries.
+ */
+int console_open(int mode)
 {
 	int f, fd;
 	int m;
@@ -444,15 +466,16 @@ int serial_open(char *dev, int mode)
 	 *	Retry the open five times.
 	 */
 	for(f = 0;f < 5; f++)
-		if ((fd = open(dev, m)) >= 0) break;
-	if (fd < 0) return(fd);
+		if ((fd = open(console_dev, m)) >= 0) break;
+
+	if (fd < 0) return fd;
 
 	/*
 	 *	Set original flags.
 	 */
 	if (m != mode)
   		fcntl(fd, F_SETFL, mode);
-	return(fd);
+	return fd;
 }
 
 /*
@@ -539,7 +562,7 @@ void set_term(int how)
 	int fd, fd2;
 	int restore_ok = 0;
 
-	if ((fd = serial_open(console_dev, O_RDWR|O_NOCTTY)) < 0) {
+	if ((fd = console_open(O_RDWR|O_NOCTTY)) < 0) {
 		log(L_VB, "can't open %s", console_dev);
 		return;
 	}
@@ -617,7 +640,7 @@ void print(char *s)
 {
 	int fd;
 
-	if ((fd = serial_open(console_dev, O_WRONLY|O_NOCTTY|O_NDELAY)) >= 0) {
+	if ((fd = console_open(O_WRONLY|O_NOCTTY|O_NDELAY)) >= 0) {
 		write(fd, s, strlen(s));
 		close(fd);
 	}
@@ -815,10 +838,11 @@ int spawn(CHILD *ch, int *res)
 			 *	of the console after exit of the leader.
 			 */
 			setsid();
-			f = serial_open(console_dev, O_RDWR|O_NOCTTY);
-			(void)ioctl(f, TIOCSCTTY, 0);
-  			dup(f);
-  			dup(f);
+			if ((f = console_open(O_RDWR|O_NOCTTY)) >= 0) {
+				(void)ioctl(f, TIOCSCTTY, 0);
+  				dup(f);
+  				dup(f);
+			}
 			if ((pid = fork()) < 0) {
   				log(L_VB, "cannot fork");
 				exit(1);
@@ -859,7 +883,7 @@ int spawn(CHILD *ch, int *res)
 			/* Set ioctl settings to default ones */
 			set_term(0);
   		} else {
-			if ((f = serial_open(console_dev, O_RDWR|O_NOCTTY)) < 0) {
+			if ((f = console_open(O_RDWR|O_NOCTTY)) < 0) {
 				log(L_VB, "open(%s): %s", console_dev,
 					strerror(errno));
 				f = open("/dev/null", O_RDWR);
@@ -1413,7 +1437,7 @@ int ask_runlevel()
 	int fd;
 
 	set_term(0);
-	fd = serial_open(console_dev, O_RDWR|O_NOCTTY);
+	fd = console_open(O_RDWR|O_NOCTTY);
 
 	if (fd < 0) return('S');
 
@@ -1674,7 +1698,7 @@ int check_pipe(int fd)
 		return 0;
 	if (read(fd, signature, 8) != 8)
 		 return 0;
-	return strcmp(Signature, signature) == 0;
+	return strncmp(Signature, signature, 8) == 0;
 }
 
 /*
@@ -1820,9 +1844,31 @@ void check_init_fifo(void)
   int n;
   fd_set fds;
   int quit = 0;
-  struct stat st;
+  struct stat st, st2;
 
-  /* Try to open /dev/initctl */
+  /*
+   *	First, try to create /dev/initctl if not present.
+   */
+  if (stat(INIT_FIFO, &st2) < 0 && errno == ENOENT)
+	(void)mkfifo(INIT_FIFO, 0600);
+
+  /*
+   *	If /dev/initctl is open, stat the file to see if it
+   *	is still the _same_ inode.
+   */
+  if (pipe_fd >= 0) {
+	fstat(pipe_fd, &st);
+	if (stat(INIT_FIFO, &st2) < 0 ||
+	    st.st_dev != st2.st_dev ||
+	    st.st_ino != st2.st_ino) {
+		close(pipe_fd);
+		pipe_fd = -1;
+	}
+  }
+
+  /*
+   *	Now finally try to open /dev/initctl
+   */
   if (pipe_fd < 0) {
 	if ((pipe_fd = open(INIT_FIFO, O_RDWR|O_NONBLOCK)) >= 0) {
 		fstat(pipe_fd, &st);
@@ -1870,7 +1916,15 @@ void check_init_fifo(void)
 		continue;
 	}
 
-	/* Process request. */
+	/*
+	 *	This is a convenient point to also try to
+	 *	find the console device or check if it changed.
+	 */
+	console_init();
+
+	/*
+	 *	Process request.
+	 */
 	if (request.magic != INIT_MAGIC || n != sizeof(request)) {
 		log(L_VB, "got bogus initrequest");
 		continue;
@@ -1902,7 +1956,9 @@ void check_init_fifo(void)
 	}
   }
 
-  /* We come here if the pipe couldn't be opened. */
+  /*
+   *	We come here if the pipe couldn't be opened.
+   */
   if (pipe_fd < 0) pause();
 
 }
@@ -2116,7 +2172,7 @@ void process_signals()
 	 *	SIGUSR1 means close and reopen /dev/initctl
 	 */
 #if DEBUG
-	log(L_VB, "got SIGHUP");
+	log(L_VB, "got SIGUSR1");
 #endif
 	close(pipe_fd);
 	pipe_fd = -1;
@@ -2133,7 +2189,6 @@ int init_main()
   CHILD *ch;
   sigset_t sgt;
   struct sigaction sa;
-  char *s;
 
   if (!reload) {
   
@@ -2191,14 +2246,14 @@ int init_main()
   siginterrupt (SIGUSR1,  1);
   siginterrupt (SIGWINCH, 1);
 
+  console_init();
+
   if (!reload) {
 
   	/* Close whatever files are open, and reset the console. */
 	close(0);
 	close(1);
 	close(2);
-	if ((s = getenv("CONSOLE")) != NULL)
-		console_dev = s;
   	set_term(0);
   	setsid();
 
@@ -2330,6 +2385,7 @@ int main(int argc, char *argv[])
 		maxproclen = strlen(argv[0]);
 		reload = 1;
 		wrote_iosave = 1;
+		setproctitle("init [%c]",runlevel);
 
 		init_main();
 	}
