@@ -11,7 +11,7 @@
  *		encrypted root password is "x" the shadow
  *		password will be used.
  *
- * Version:	@(#)sulogin 2.74 10-May-1998 miquels@cistron.nl
+ * Version:	@(#)sulogin 2.76 20 Nov-1998 miquels@cistron.nl
  *
  */
 
@@ -37,9 +37,10 @@
 #define SHADOW	"/etc/shadow"
 #define BINSH	"/bin/sh"
 
-char *Version = "@(#)sulogin 2.74 10-May-1998 miquels@cistron.nl";
+char *Version = "@(#)sulogin 2.76 20-Nov-1998 miquels@cistron.nl";
 
 int timeout = 0;
+int profile = 0;
 
 #if 0
 /*
@@ -105,7 +106,7 @@ int valid(char *pass)
 	 *	3 bytes for the signature $1$
 	 *	up to 8 bytes for the salt
 	 *	$
-	 *	the MD5 hash (128 bits or 16 bytes) encoded in base64 = 24 bytes
+	 *	the MD5 hash (128 bits or 16 bytes) encoded in base64 = 22 bytes
 	 */
 	if (strncmp(pass, "$1$", 3) == 0) {
 		for(s = pass + 3; *s && *s != '$'; s++)
@@ -231,7 +232,7 @@ struct passwd *getrootpwent()
 char *getpasswd()
 {
 	struct termios old, tty;
-	static char pass[16];
+	static char pass[128];
 	char *ret = pass;
 	int i;
 
@@ -245,12 +246,12 @@ char *getpasswd()
 	tty.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|TOSTOP);
 	tcsetattr(0, TCSANOW, &tty);
 
-	pass[15] = 0;
+	pass[sizeof(pass) - 1] = 0;
 	if (timeout) alarm(timeout);
-	if (read(0, pass, 15) <= 0)
+	if (read(0, pass, sizeof(pass) - 1) <= 0)
 		ret = NULL;
 	else {
-		for(i = 0; i < 15; i++)
+		for(i = 0; i < sizeof(pass) && pass[i]; i++)
 			if (pass[i] == '\r' || pass[i] == '\n') {
 				pass[i] = 0;
 				break;
@@ -268,8 +269,8 @@ char *getpasswd()
  */
 void sushell(struct passwd *pwd)
 {
-	char shell[64];
-	char home[64];
+	char shell[128];
+	char home[128];
 	char *p;
 	char *sushell;
 
@@ -291,25 +292,25 @@ void sushell(struct passwd *pwd)
 		p = sushell;
 	else
 		p++;
-	strcpy(shell, "-");
-	strcat(shell, p);
+	snprintf(shell, sizeof(shell), profile ? "-%s" : "%s", p);
 
 	/*
 	 *	Set some important environment variables.
 	 */
-	getcwd(home, 64);
+	getcwd(home, sizeof(home));
 	setenv("HOME", home, 1);
 	setenv("LOGNAME", "root", 1);
 	setenv("USER", "root", 1);
-	setenv("SHELL", sushell, 1);
 
 	/*
 	 *	Try to execute a shell.
 	 */
+	setenv("SHELL", sushell, 1);
 	execl(sushell, shell, NULL);
 	perror(sushell);
+
 	setenv("SHELL", BINSH, 1);
-	execl(BINSH, "-sh", NULL);
+	execl(BINSH, profile ? "-sh" : "sh", NULL);
 	perror(BINSH);
 }
 
@@ -330,10 +331,13 @@ int main(int argc, char **argv)
 	 *	See if we have a timeout flag.
 	 */
 	opterr = 0;
-	while((c = getopt(argc, argv, "t:")) != EOF) switch(c) {
+	while((c = getopt(argc, argv, "pt:")) != EOF) switch(c) {
 		case 't':
 			timeout = atoi(optarg);
 			break;
+		case 'p':
+			profile = 1;
+			break;
 		default:
 			usage();
 			/* Do not exit! */
@@ -400,7 +404,7 @@ int main(int argc, char **argv)
 	}
 
 	/*
-	 *	User gave Control-D.
+	 *	User pressed Control-D.
 	 */
 	return 0;
 }
